# -*- coding: us-ascii -*-
from collective.searchandreplace.interfaces import ISearchReplaceUtility
from zope.component import getUtility
from zope.publisher.browser import BrowserView


class SearchReplaceTable(BrowserView):
    """ View class for search and replace preivew table widget."""

    def getItems(self):
        """ Get preview items """
        # Set search parameters
        srutil = getUtility(ISearchReplaceUtility)
        stext = None
        if 'form.findWhat' in self.request:
            stext = self.request['form.findWhat']
        if not stext:
            return []
        if 'form.searchSubfolders' in self.request:
            subfolders = True
        else:
            subfolders = False
        if 'form.matchCase' in self.request:
            mcase = True
        else:
            mcase = False
        # Get search results
        results = srutil.searchObjects(
            self.context,
            stext,
            searchSubFolders=subfolders,
            matchCase=mcase)
        return results

    def getRelativePath(self, path):
        """ Get a relative path """
        cpath = '/'.join(self.context.getPhysicalPath())
        rpath = path[len(cpath):]
        if rpath:
            rpath = '.' + rpath
        else:
            rpath = './'
        return rpath



from collective.searchandreplace.interfaces import ISearchReplaceUtility
from collective.searchandreplace.tests.base import SearchAndReplaceTestCase
from zope.component import getUtility


class testMatchCase(SearchAndReplaceTestCase):
    """ Ensure that the match case flag works   """

    def afterSetUp(self):
        self.srutil = getUtility(ISearchReplaceUtility)

    def testNoMatchCase(self):
        self.setRoles(['Manager'])
        self.portal.invokeFactory('Document', 'doc1')
        doc1 = getattr(self.portal, 'doc1')
        doc1.setTitle('Test Title')
        doc1.setText('Test Case')
        results = self.srutil.searchObjects(
            doc1,
            'test case',
            replaceText='foo',
            matchCase=False)
        self.assertEqual(len(results), 1)

    def testMatchCase(self):
        self.setRoles(['Manager'])
        self.portal.invokeFactory('Document', 'doc2')
        doc2 = getattr(self.portal, 'doc2')
        doc2.setTitle('test title')
        doc2.setText('Test Case')
        results = self.srutil.searchObjects(
            doc2,
            'test case',
            replaceText='foo',
            matchCase=True)
        self.assertEqual(len(results), 0)

    def testOneMatch(self):
        self.setRoles(['Manager'])
        self.portal.invokeFactory('Document', 'doc1')
        doc1 = getattr(self.portal, 'doc1')
        doc1.setTitle('Test Title')
        doc1.setText('Test Case')
        self.portal.invokeFactory('Document', 'doc2')
        doc2 = getattr(self.portal, 'doc2')
        doc2.setTitle('test title')
        doc2.setText('test case')
        # path1 = '%s' %  '/'.join(doc1.getPhysicalPath())
        # path2 = '%s' % '/'.join(doc2.getPhysicalPath())
        # paths = {path1: {'body': [0]}, path2: {'body': [0]}}
        results = self.srutil.searchObjects(
            self.portal,
            'test case',
            replaceText='foo',
            matchCase=True,
            # searchItems=paths
        )
        self.assertEqual(len(results), 1)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testMatchCase))
    return suite



from Products.PloneTestCase.PloneTestCase import default_user
from collective.searchandreplace.interfaces import ISearchReplaceUtility
from collective.searchandreplace.tests.base import SearchAndReplaceTestCase
from zope.component import getUtility


class testReplaceWhere(SearchAndReplaceTestCase):
    """Ensure that we can replace in title, description and text."""

    def afterSetUp(self):
        self.srutil = getUtility(ISearchReplaceUtility)

    def testReplaceText(self):
        self.setRoles(['Manager'])
        self.portal.invokeFactory('Document', 'doc1')
        doc1 = getattr(self.portal, 'doc1')
        doc1.setTitle('Test Title')
        doc1.setDescription('Test Description')
        doc1.setText('Test Case')
        results = self.srutil.searchObjects(
            doc1,
            'test case',
            replaceText='foo',
            matchCase=False)
        self.assertEqual(len(results), 1)

    def testReplaceTitle(self):
        self.setRoles(['Manager'])
        self.portal.invokeFactory('Document', 'doc2')
        doc1 = getattr(self.portal, 'doc2')
        doc1.setTitle('Test Title')
        doc1.setDescription('Test Description')
        doc1.setText('Test Case')
        results = self.srutil.searchObjects(
            doc1,
            'test title',
            replaceText='foo',
            matchCase=False)
        self.assertEqual(len(results), 1)

    def testReplaceDescription(self):
        self.setRoles(['Manager'])
        self.portal.invokeFactory('Document', 'doc1')
        doc1 = getattr(self.portal, 'doc1')
        doc1.setTitle('Test Title')
        doc1.setDescription('Test Description')
        doc1.setText('Test Case')
        results = self.srutil.searchObjects(
            doc1,
            'test desc',
            replaceText='foo',
            matchCase=False)
        self.assertEqual(len(results), 1)

    def testReplaceOnlyEditableContent(self):
        self.loginAsPortalOwner()
        # /mainfolder
        self.portal.invokeFactory('Folder', 'mainfolder')
        mainfolder = getattr(self.portal, 'mainfolder')
        mainfolder.setTitle('Test Title')

        # /mainfolder/maindoc
        mainfolder.invokeFactory('Document', 'maindoc')
        maindoc = getattr(mainfolder, 'maindoc')
        maindoc.setTitle('Test Title')

        # /mainfolder/subfolder
        mainfolder.invokeFactory('Folder', 'subfolder')
        subfolder = getattr(mainfolder, 'subfolder')
        subfolder.setTitle('Test Title')

        # /mainfolder/subfolder/subdoc
        subfolder.invokeFactory('Document', 'subdoc')
        subdoc = getattr(subfolder, 'subdoc')
        subdoc.setTitle('Test Title')

        # Give test user a local Editor role on the sub folder.
        subfolder.manage_addLocalRoles(default_user, ('Editor',))

        self.portal.portal_catalog.reindexObject(mainfolder)
        self.portal.portal_catalog.reindexObject(maindoc)
        self.portal.portal_catalog.reindexObject(subfolder)
        self.portal.portal_catalog.reindexObject(subdoc)

        # We are logged in as portal owner, so we can edit everything.
        results = self.srutil.searchObjects(
            mainfolder,
            'test title',
            replaceText='foo',
            matchCase=False)
        self.assertEqual(len(results), 4)

        # Login as the test user again.
        self.logout()
        self.login()
        # Now we can edit less: only the sub folder and sub doc.
        results = self.srutil.searchObjects(
            mainfolder,
            'test title',
            replaceText='foo',
            matchCase=False)
        self.assertEqual(len(results), 2)
        paths = [x['path'] for x in results]
        self.assertIn('/'.join(subfolder.getPhysicalPath()), paths)
        self.assertIn('/'.join(subdoc.getPhysicalPath()), paths)
        self.assertNotIn('/'.join(mainfolder.getPhysicalPath()), paths)
        self.assertNotIn('/'.join(maindoc.getPhysicalPath()), paths)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testReplaceWhere))
    return suite



import doctest
from Products.Five import fiveconfigure
from Products.Five import zcml
from Products.PloneTestCase.PloneTestCase import PloneTestCase, FunctionalTestCase
from Products.PloneTestCase.PloneTestCase import setupPloneSite
from Products.PloneTestCase.layer import onsetup
from Testing import ZopeTestCase as ztc


@onsetup
def setup_searchandreplace_project():
    """
    Load and install packages required for the collective.searchandreplace tests
    """

    fiveconfigure.debug_mode = True

    import collective.searchandreplace
    zcml.load_config('configure.zcml', collective.searchandreplace)

    fiveconfigure.debug_mode = False

    ztc.installPackage('collective.searchandreplace')


setup_searchandreplace_project()
setupPloneSite(
    with_default_memberarea=0,
    extension_profiles=['collective.searchandreplace:default'])

oflags = (doctest.ELLIPSIS |
          doctest.NORMALIZE_WHITESPACE)

prod = "collective.searchandreplace"


class SearchAndReplaceTestCase(PloneTestCase):
    """ Test Class """

    def _setupHomeFolder(self):
        """ Ugly hack to keep the underlying testing framework from trying to create a user folder."""
        pass


class SearchAndReplaceFunctionalTestCase(FunctionalTestCase):
    """ Functional test class """



# -*- coding: us-ascii -*-
from Acquisition import aq_parent
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from collective.searchandreplace import SearchAndReplaceMessageFactory as _
from collective.searchandreplace.interfaces import ISearchReplaceUtility
from collective.searchandreplace.browser.customwidgets import \
     TwoLineTextAreaWidget
from five.formlib.formbase import AddForm
from plone.app.layout.navigation.defaultpage import isDefaultPage
from zope.component import getUtility
from zope.formlib.form import FormFields, action
from zope.interface import Interface
from zope.schema import Text, Bool


def validate_searchreplaceform(form, action, data):
    """ Validate the Search/Replace form. """


class ISearchReplaceForm(Interface):
    """ Interface for Search and Replace form """

    findWhat = Text(title=_(u'Find What'),
                    description=_(u'Enter the text to find.'),
                    required=True)

    replaceWith = Text(title=_(u'Replace With'),
                       description=_(u'Enter the text to replace the '
                                     'original text with.'),
                       required=False)

    searchSubfolders = Bool(title=_(u'Search Subfolders'),
                            description=_(u'If checked, this will '
                                          'recursively search through '
                                          'any selected folders and '
                                          'their children, replacing at '
                                          'each level.'),
                            default=True,
                            required=True)

    matchCase = Bool(title=_(u'Match Case'),
                     description=_(
                         u'Check the box for a case sensitive search.'),
                     default=False,
                     required=True)


class SearchReplaceForm(AddForm):
    """ """

    @property
    def form_fields(self):
        form_fields = FormFields(ISearchReplaceForm)
        container = aq_parent(self.context)
        if not self.context.isPrincipiaFolderish and not isDefaultPage(container, self.context):
            form_fields = form_fields.omit('searchSubfolders')
        form_fields['findWhat'].custom_widget = TwoLineTextAreaWidget
        form_fields['replaceWith'].custom_widget = TwoLineTextAreaWidget
        return form_fields

    label = _(u'Search and Replace')
    description = _(u'Search and replace text found in documents.')
    template = ViewPageTemplateFile('pageform.pt')

    @action(_(u'Preview'),
            validator=None,
            name=u'Preview')
    def action_preview(self, action, data):
        """ Preview files to be changed. """
        self.form_reset = False

    @action(_(u'Replace'),
            validator=validate_searchreplaceform,
            name=u'Replace')
    def action_replace(self, action, data):
        """ Replace text for all files. """
        srutil = getUtility(ISearchReplaceUtility)
        if 'form.affectedContent' in self.request:
            # Do only the selected items
            # nitems = len(self.request['form.affectedContent'])
            items = srutil.parseItems(self.request['form.affectedContent'])
            nitems = 0
            for page_url, page_result in items.items():
                for field, indexes in page_result.items():
                    nitems += len(indexes)
            replaced = srutil.searchObjects(
                self.context,
                data['findWhat'],
                searchSubFolders=data.get('searchSubfolders', False),
                matchCase=data['matchCase'],
                replaceText=data['replaceWith'],
                doReplace=True,
                searchItems=items)
            IStatusMessage(self.request).addStatusMessage(
                _(u'Search text replaced in ${replaced} of ${items} '
                  'instance(s).',
                  mapping={'replaced': replaced, 'items': nitems}),
                type='info')
        else:
            # Do everything you can find
            replaced = srutil.searchObjects(
                self.context,
                data['findWhat'],
                searchSubFolders=data.get('searchSubfolders', False),
                matchCase=data['matchCase'],
                replaceText=data['replaceWith'],
                doReplace=True)
            IStatusMessage(self.request).addStatusMessage(
                _(u'Search text replaced in all ${items} instance(s).',
                  mapping={'items': replaced}),
                type='info')
        self.request.response.redirect(self.context.absolute_url())
        return ''

    @action(_(u'Reset'),
            validator=None,
            name=u'Reset')
    def action_reset(self, action, data):
        """ Reset the form fields to their defaults. """



from Testing.ZopeTestCase import FunctionalDocFileSuite
from collective.searchandreplace.tests.base import \
    SearchAndReplaceFunctionalTestCase
from collective.searchandreplace.tests.base import oflags
from collective.searchandreplace.tests.base import prod
from unittest import TestSuite


def test_suite():
    suite = TestSuite()

    basicsearchtest = FunctionalDocFileSuite('tests/basicsearch.txt',
                                             package=prod,
                                             test_class=SearchAndReplaceFunctionalTestCase,
                                             optionflags=oflags)

    searchavailabletest = FunctionalDocFileSuite('tests/searchavailable.txt',
                                                 package=prod,
                                                 test_class=SearchAndReplaceFunctionalTestCase,
                                                 optionflags=oflags)

    suite.addTests((basicsearchtest, searchavailabletest))

    return suite



from zope.interface import Interface


class ISearchReplaceable(Interface):
    """ Marker interface for searching and replacing """


class ISearchReplaceUtility(Interface):
    """ Search and replace utility interface """

    def getItems():
        """ Get items for preview. """



PROFILE_ID = 'profile-collective.searchandreplace:default'


def dummy_upgrade_step(context):
    return


def run_actions_step(context):
    context.runImportStepFromProfile(PROFILE_ID, 'actions')


def run_rolemap_step(context):
    context.runImportStepFromProfile(PROFILE_ID, 'rolemap')



##################################################################################
#    Copyright (c) 2009 Novell, All rights reserved.
#    Portions copyright 2009 Massachusetts Institute of Technology, All rights reserved.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 2.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##################################################################################

__author__ = '''Brent Lambert, David Ray, Jon Thomas'''
__version__ = '$ Revision 0.0 $'[11:-2]

from setuptools import setup, find_packages

version = '5.1'

setup(name='collective.searchandreplace',
      version=version,
      description="Batch Search and Replace",
      long_description=(open("README.rst").read() + "\n" +
                        open("CHANGES.rst").read()),
      # Get more strings from
      # https://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          "Framework :: Plone",
          "Framework :: Plone :: 4.3",
          "Framework :: Plone :: 5.0",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Topic :: Software Development :: Libraries :: Python Modules",
          ],
      keywords='batch search replace',
      author='enPraxis',
      author_email='info@enpraxis.net',
      maintainer='Maurits van Rees',
      maintainer_email='m.van.rees@zestsoftware.nl',
      url='https://github.com/collective/collective.searchandreplace',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'Acquisition',
          'five.formlib',
          'plone.app.layout',
          'plone.app.textfield',
          'Products.CMFCore',
          'Products.CMFPlone',
          'Products.statusmessages',
          'setuptools',
          'Zope2',
          'zope.component',
          'zope.formlib',
          'zope.i18nmessageid',
          'zope.interface',
          'zope.publisher',
          'zope.schema',
      ],
      extras_require={
          'test': [
              'Products.PloneTestCase',
          ],
      },
      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



# -*- coding: us-ascii -*-

import re
from Acquisition import aq_parent
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName
from plone.app.layout.navigation.defaultpage import isDefaultPage
from plone.app.textfield import RichTextValue

searchflags = re.DOTALL | re.UNICODE | re.MULTILINE


def _to_unicode(s):
    assert isinstance(s, basestring)
    if not isinstance(s, unicode):
        s = s.decode('utf-8')
    return s


class SearchReplaceUtility(object):
    """ Search and replace utility. """

    def searchObjects(self, context, find, **kwargs):
        """ Search objects and optionally do a replace. """
        # Get search parameters
        cpath = context.getPhysicalPath()
        if 'searchSubFolders' in kwargs:
            ssf = kwargs['searchSubFolders']
        else:
            ssf = True
        if 'matchCase' in kwargs:
            mc = kwargs['matchCase']
        else:
            mc = False
        if 'replaceText' in kwargs:
            rtext = kwargs['replaceText']
        else:
            rtext = None
        if 'doReplace' in kwargs:
            replace = kwargs['doReplace']
        else:
            replace = False
        if 'searchItems' in kwargs:
            sitems = kwargs['searchItems']
        else:
            sitems = None
        # Get Regex matcher
        sflags = mc and searchflags or (searchflags | re.IGNORECASE)
        matcher = re.compile(find, sflags)
        # Get items to search
        query = {'query': '/'.join(cpath)}
        if context.isPrincipiaFolderish and not ssf:
            query['depth'] = 1
        container = aq_parent(context)
        if isDefaultPage(container, context) and ssf:
            query['query'] = '/'.join(container.getPhysicalPath())
        catalog = getToolByName(context, 'portal_catalog')
        brains = catalog(path=query)
        memship = getToolByName(context, 'portal_membership')
        checkPermission = memship.checkPermission
        # Match objects
        results = []
        replaced = 0
        for b in brains:
            ipath = b.getPath()
            if not sitems or ipath in sitems:
                obj = b.getObject()
                # Does the user have the modify permission on this object?
                if not checkPermission(ModifyPortalContent, obj):
                    continue
                # If there is a filtered list of items, and it
                # is in the list, or if there is no filter
                # then process the item
                if replace and rtext:
                    # Do a replace
                    if sitems:
                        sitem = sitems[ipath]
                    else:
                        sitem = None
                    rep = self._replaceObject(matcher,
                                              obj,
                                              cpath,
                                              rtext,
                                              sitem)
                    replaced += rep
                elif not replace:
                    # Just find the matches and return info
                    result = self._searchObject(matcher, obj)
                    if result:
                        results += result
        if replace:
            return replaced
        else:
            return results

    def _replaceObject(self, matcher, obj, cpath, rtext, mobjs):
        """ Replace text in objects """
        replaced = 0
        # rtext is already unicode
        if mobjs:
            # Replace only the objects specified in mobjs
            if 'title' in mobjs:
                title = _to_unicode(obj.aq_base.Title())
                result = self._replaceText(matcher,
                                           title,
                                           rtext,
                                           mobjs['title'])
                if result[0]:
                    replaced += result[0]
                    obj.aq_base.setTitle(result[1])
            if 'description' in mobjs:
                desc = self._getDesc(obj)
                if desc:
                    result = self._replaceText(matcher,
                                               desc,
                                               rtext,
                                               mobjs['description'])
                    if result[0]:
                        replaced += result[0]
                        obj.aq_base.setDescription(result[1])
            if 'body' in mobjs:
                body = self._getRawText(obj)
                if body:
                    result = self._replaceText(matcher,
                                               body,
                                               rtext,
                                               mobjs['body'])
                    if result[0]:
                        replaced += result[0]
                        self._setText(obj, result[1])
        else:
            # Replace all occurences
            title = _to_unicode(obj.aq_base.Title())
            result = self._replaceText(matcher,
                                       title,
                                       rtext,
                                       None)
            if result[0]:
                replaced += result[0]
                obj.aq_base.setTitle(result[1])
            desc = self._getDesc(obj)
            if desc:
                result = self._replaceText(matcher,
                                           desc,
                                           rtext,
                                           None)
                if result[0]:
                    replaced += result[0]
                    obj.setDescription(result[1])
            body = self._getRawText(obj)
            if body:
                result = self._replaceText(matcher,
                                           body,
                                           rtext,
                                           None)
                if result[0]:
                    replaced += result[0]
                    self._setText(obj, result[1])
        # don't have to utf-8 encoding
        if replaced:
            obj.reindexObject()
        return replaced

    def _replaceText(self, matcher, text, rtext, indexes):
        """ Replace instances """
        newtext = ''
        mindex = 0
        replaced = 0
        mobj = matcher.finditer(text)
        for x in mobj:
            start, end = x.span()
            if not indexes or start in indexes:
                newtext += text[mindex:start]
                newtext += rtext
                mindex = end
                replaced += 1
        newtext += text[mindex:]
        return replaced, newtext

    def _searchObject(self, matcher, obj):
        """ Find location of search strings """
        results = []
        path = '/'.join(obj.getPhysicalPath())
        title = _to_unicode(obj.aq_base.Title())
        mobj = matcher.finditer(title)
        for x in mobj:
            start, end = x.span()
            results.append({
                'path': path,
                'url': obj.absolute_url(),
                'line': 'title',
                'pos': '%d' % start,
                'text': self._getLinePreview(title,
                                             start,
                                             end), })
        desc = self._getDesc(obj)
        if desc:
            mobj = matcher.finditer(desc)
            for x in mobj:
                start, end = x.span()
                results.append({
                    'path': path,
                    'url': obj.absolute_url(),
                    'line': 'description',
                    'pos': '%d' % start,
                    'text': self._getLinePreview(desc,
                                                 start,
                                                 end), })
        text = self._getRawText(obj)
        if text:
            mobj = matcher.finditer(text)
            for x in mobj:
                start, end = x.span()
                results.append({
                    'path': path,
                    'url': obj.absolute_url(),
                    'line': '%d' % self._getLineNumber(text,
                                                       start),
                    'pos': '%d' % start,
                    'text': self._getLinePreview(text,
                                                 start,
                                                 end), })
        return results

    def _getDesc(self, obj):
        if hasattr(obj.aq_base, 'getRawDescription'):
            desc = obj.aq_base.getRawDescription()
        else:
            desc = getattr(obj.aq_base, 'description', u'')
        return _to_unicode(desc)

    def _getRawText(self, obj):
        text = None
        if hasattr(obj, 'getText'):
            baseunit = obj.getField('text').getRaw(obj, raw=True)
            if isinstance(baseunit.raw, unicode):
                text = baseunit.raw
            else:
                text = _to_unicode(obj.aq_base.getRawText())
        elif hasattr(obj, 'text') and hasattr(obj.text, 'raw'):
            text = _to_unicode(obj.text.raw)
        return text

    def _setText(self, obj, text):
        obj = obj.aq_base
        if hasattr(obj, 'setText'):
            obj.setText(text)
        else:
            obj.text = RichTextValue(
                text, obj.text.mimeType, obj.text.outputMimeType)

    def _getLineNumber(self, text, index):
        return text.count('\n', 0, index) + 1

    def _getLinePreview(self, text, start, end):
        sindex = text[:start].rfind('\n')
        if -1 == sindex:
            sindex = 0
        eindex = text[end:].find('\n')
        if -1 == eindex:
            eindex = None
        else:
            eindex += end
        return (text[sindex:start],
                text[start:end],
                text[end:eindex])

    def parseItems(self, items):
        """ Get list of items from form values """
        itemd = {}
        if not isinstance([], type(items)):
            items = [items]
        for x in items:
            try:
                line, pos, path = x.split(':')
            except ValueError:
                break
            if path not in itemd:
                itemd[path] = {}
            if 'title' == line:
                if 'title' not in itemd[path]:
                    itemd[path]['title'] = []
                itemd[path]['title'].append(int(pos))
            elif 'description' == line:
                if 'description' not in itemd[path]:
                    itemd[path]['description'] = []
                itemd[path]['description'].append(int(pos))
            else:
                if 'body' not in itemd[path]:
                    itemd[path]['body'] = []
                itemd[path]['body'].append(int(pos))
        return itemd



from zope.i18nmessageid import MessageFactory

# Define permission string, so we can import it if needed.  See also
# the permission in configure.zcml.
UseSearchAndReplace = "collective.searchandreplace: Use Search And Replace"
# Message factory:
SearchAndReplaceMessageFactory = MessageFactory('SearchAndReplace')


def initialize(context):
    """Initializer called when used as a Zope 2 product."""
